module LinearGWPO where

import Data.List
import Term
import Rule
import TRS
import SMT
import Polynomial

-- f_0 + f_1 x_1 + ... + f_n x_n
type Coefficients = (Int,[Int])

-- f(x1,...,xn) = f_0 + f_1 x_1 + ... + f_n x_n
type Interpretation = (String,Coefficients)

type Algebra = [Interpretation]

type Precedence = [(String, Int)]

type Parameter = (Algebra, Precedence)

data Order = Linear Parameter

type Table = [(Term,Term)]

showMonomial :: (Int, String) -> String
showMonomial (0, _) = "0"
showMonomial (1, x) = x
showMonomial (n, x) = show n ++ " " ++ x

-- print as polynomials
showCoefficients :: Coefficients -> String
showCoefficients (a0, as) = 
  case ms1 ++ ms2 of
    [] -> "0"
    ms -> intercalate " + " ms
  where
    ms1 = [ showMonomial (ai, "x" ++ show i) 
          | (i, ai) <- zip [1 :: Int ..] as, ai /= 0 ]
    ms2 = if a0 == 0 then [] else [show a0]

showInterpretation :: Interpretation -> String
showInterpretation (f, (a0,[])) =
  f ++ "_A = " ++ show a0
showInterpretation (f, p@(_,as)) =
  f ++ "_A(" ++ xs ++ ") = " ++ showCoefficients p
  where
    xs = intercalate "," [ "x" ++ show i | i <- [1 .. length as] ]

showAlgebra :: Algebra -> String
showAlgebra interpretations =
  "linear polynomial interpretations on N:\n" ++
  "\n" ++
  unlines 
    [ "  " ++ showInterpretation interpretation 
    | interpretation <- interpretations ] ++ "\n"

showPrecedence' :: Precedence -> String
showPrecedence' []  = error "showPrecedence"
showPrecedence' [(f,_)] = f
showPrecedence' ((f,m) : p@((_,n) : _))
  | m == n = f ++ " = " ++ showPrecedence' p
  | m >  n = f ++ " > " ++ showPrecedence' p
  | otherwise = error "showPrecedence"

showPrecedence :: Precedence -> String
showPrecedence []  =  
  "precedence:\n\n" ++ "    empty\n\n"
showPrecedence [_] = 
  "precedence:\n\n" ++ "    empty\n\n"
showPrecedence a = 
  "precedence:\n\n" ++
  "  " ++ showPrecedence' (sortBy f a) ++ "\n\n"
  where f (_,m) (_,n) = compare n m

showParameter :: Parameter -> String
showParameter (a, p) =
  showAlgebra a ++
  showPrecedence p

instance Show Order where
  show (Linear param) = showParameter param

interpret :: Algebra -> Term -> Polynomial
interpret _ (V x) = var x 
interpret a (F f ts)
  | Just (a0, as) <- lookup f a = 
    Polynomial.plus (Polynomial.num a0 : [ Polynomial.times ai (interpret a ti) | (ai, ti) <- zip as ts ])
interpret _ _ = error "interpret"

-- >=_A
geqA :: Algebra -> Rule -> Bool
geqA a (l, r) = Polynomial.geq (interpret a l) (interpret a r)

-- >_A
gtA :: Algebra -> Rule -> Bool
gtA a (l, r) = Polynomial.gt (interpret a l) (interpret a r)

-- f >= g
geqPrec :: Precedence -> Rule -> Bool
geqPrec p (F f _, F g _)
  | m <- lookup f p, n <- lookup g p = m >= n
geqPrec _ _ = error "geqPrec"

-- f > g
gtPrec :: Precedence -> Rule -> Bool
gtPrec p (F f _, F g _)
  | m <- lookup f p, n <- lookup g p = m > n
gtPrec _ _ = error "geqPrec"

-- s \sqsupsetsim t
geqB :: Parameter -> Rule -> Bool
geqB (a,p) rule = 
  gtA a (Rule.mark rule) ||
  (geqA a (Rule.mark rule) && geqPrec p rule)

-- s \sqsupset t
gtB :: Parameter -> Rule -> Bool
gtB (a,p) rule = 
  gtA a (Rule.mark rule) ||
  (geqA a (Rule.mark rule) && gtPrec p rule)

geqSPO :: Parameter -> Rule -> Bool
geqSPO _ (V x, V y) = x == y
geqSPO _ (V _, _) = False
geqSPO param (F _ ss, t)
  | any (\si -> geqSPO param (si, t)) ss = True
geqSPO _ (F _ _, V _) = False
geqSPO param (s@(F _ ss), t@(F _ ts)) =
  all (\ti -> gtSPO param (s, ti)) ts &&
  (gtB param (s, t) || 
   (geqB param (s, t) && geqSPOLex param ss ts))

geqSPOLex :: Parameter -> [Term] -> [Term] -> Bool
geqSPOLex _ _  []      = True
geqSPOLex _ [] (_ : _) = False
geqSPOLex param (s : ss) (t : ts) =
  gtSPO param (s, t) || 
  (geqSPO param (s, t) && geqSPOLex param ss ts)

gtSPO :: Parameter -> Rule -> Bool
gtSPO _ (V _, _) = False
gtSPO param (F _ ss, t)
  | any (\si -> geqSPO param (si, t)) ss = True
gtSPO _ (F _ _, V _) = False
gtSPO param (s@(F _ ss), t@(F _ ts)) =
  all (\ti -> gtSPO param (s, ti)) ts &&
  (gtB param (s, t) || 
   (geqB param (s, t) && gtSPOLex param ss ts))

gtSPOLex :: Parameter -> [Term] -> [Term] -> Bool
gtSPOLex _ [] _        = False
gtSPOLex _ (_ : _)  [] = True
gtSPOLex param (s : ss) (t : ts) =
  gtSPO param (s, t) || 
  (geqSPO param (s, t) && gtSPOLex param ss ts)

gtSPOMul :: Parameter -> [Term] -> [Term] -> Bool
gtSPOMul param ss ts =
  ss'/= [] &&
  and [ or [ gtSPO param (s, t) | s <- ss' ] | t <- ts' ]
  where 
    ss' = ss \\ ts
    ts' = ts \\ ss

geqMSPO :: Parameter -> Rule -> Bool
geqMSPO param@(a,_) rule = geqA a rule && geqSPO param rule

gtMSPO :: Parameter -> Rule -> Bool
gtMSPO param@(a,_) rule = geqA a rule && gtSPO param rule

-- Encoding functions.

varF0 :: String -> Exp
varF0 f = Var ("a0_" ++ f) 

varFi :: String -> Int -> Formula
varFi f i = FVar ("a_" ++ show i ++ "_" ++ f) 

varPrec :: String -> Exp
varPrec f = Var ("p_" ++ f)

varGeqSPO :: Table -> Rule -> Formula
varGeqSPO table rule
  | Just i <- elemIndex rule table = FVar ("geqSPO_" ++ show i)
  | otherwise                      = error "varGeqSPO"

varGtSPO :: Table -> Rule -> Formula
varGtSPO table rule
  | Just i <- elemIndex rule table = FVar ("gtSPO_" ++ show i)
  | otherwise                      = error "varGtSPO"

constantPart :: Term -> Exp
constantPart (V _)    = Val 0
constantPart (F f ts) =
  SMT.plus (varF0 f : [ times01 (varFi f i) (constantPart ti)
                      | (i,ti) <- zip [1..] ts ])

coefficient :: String -> Term -> Exp
coefficient x (V y)
  | x == y    = Val 1
  | otherwise = Val 0
coefficient x (F f ts) =
  SMT.plus [ times01 (varFi f i) (LinearGWPO.coefficient x ti) | (i, ti) <- zip [1..] ts ]

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA rule@(l, r) =
  conj (SMT.geq (constantPart l) (constantPart r) :
        [ SMT.geq (LinearGWPO.coefficient x l) (LinearGWPO.coefficient x r) 
        | x <- Rule.variables rule ])

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA rule@(l, r) =
  conj (SMT.gt (constantPart l) (constantPart r) :
        [ SMT.geq (LinearGWPO.coefficient x l) (LinearGWPO.coefficient x r) 
        | x <- Rule.variables rule ])

-- f >= g
encodeGeqPrec :: Rule -> Formula
encodeGeqPrec (F f _, F g _) = SMT.geq (varPrec f) (varPrec g)
encodeGeqPrec _ = error "encodeGeqPrec"

-- f > g
encodeGtPrec :: Rule -> Formula
encodeGtPrec (F f _, F g _) = SMT.gt (varPrec f) (varPrec g)
encodeGtPrec _ = error "encodeGtPrec"

-- f(s1,...,sm) >=_B g(t1,...,tn) if 
--  - s^# >_A t^#, or
--  - s^# >_A t^# and f >= g
encodeGeqB :: Rule -> Formula
encodeGeqB rule = 
  disj [encodeGtA (Rule.mark rule),
        conj [encodeGeqA (Rule.mark rule), encodeGeqPrec rule]]

-- f(s1,...,sm) >_B g(t1,...,tn) if 
--  - s^# >_A t^#, or
--  - s^# >=_A t^# and f >= g
encodeGtB :: Rule -> Formula
encodeGtB rule = 
  disj [encodeGtA (Rule.mark rule),
        conj [encodeGeqA (Rule.mark rule), encodeGtPrec rule]]

-- s >=_spo t
encodeGeqSPO :: Table -> Rule -> Formula
encodeGeqSPO _ (V x, V y)
  | x == y    = And []
encodeGeqSPO _ (V _, _) = Or []
encodeGeqSPO table rule@(F _ _, V _) =
  encodeGeqSPO1 table rule
encodeGeqSPO table rule@(F _ _, F _ _) =
  disj [encodeGeqSPO1 table rule, encodeGeqSPO23 table rule]

encodeGeqSPO1 :: Table -> Rule -> Formula
encodeGeqSPO1 table (F _ ss, t) =
  disj [ varGeqSPO table (si, t) | si <- ss ]
encodeGeqSPO1 _ _ = error "encodeGeqSPO1"

encodeGeqSPO23 :: Table -> Rule -> Formula
encodeGeqSPO23 table (s@(F _ ss), t@(F _ ts)) =
  conj [conj [ varGtSPO table (s, ti) | ti <- ts ],
        disj [encodeGtB (s, t), 
              conj [encodeGeqB (s, t), encodeGeqSPOLex table ss ts]]]
encodeGeqSPO23 _ _ = error "encodeGeqSPO2"

-- ss >=_spo^lex ts
encodeGeqSPOLex :: Table -> [Term] -> [Term] -> Formula
encodeGeqSPOLex _ _  []      = And []
encodeGeqSPOLex _ [] (_ : _) = Or []
encodeGeqSPOLex table (s : ss) (t : ts) =
  disj [varGtSPO table (s, t),
        conj [varGeqSPO table (s, t), encodeGeqSPOLex table ss ts]]

-- ss >=_spo^mul ts
encodeGeqSPOMul :: Table -> [Term] -> [Term] -> Formula
encodeGeqSPOMul table ss ts =
  conj [ disj [ encodeGeqSPO table (s, t) | s <- ss \\ ts ] 
       | t <- ts \\ ss ]

-- s >_spo t
encodeGtSPO :: Table -> Rule -> Formula
encodeGtSPO _ (V _, _) = Or []
encodeGtSPO table (s@(F _ _), t@(V _)) =
  encodeGtSPO1 table (s, t)
encodeGtSPO table rule@(F _ _, F _ _) =
  disj [encodeGtSPO1 table rule, encodeGtSPO23 table rule]

encodeGtSPO1 :: Table -> Rule -> Formula
encodeGtSPO1 table (F _ ss, t) =
  disj [ varGeqSPO table (si, t) | si <- ss ]
encodeGtSPO1 _ _ = error "encodeGtSPO1"

encodeGtSPO23 :: Table -> Rule -> Formula
encodeGtSPO23 table (s@(F _ ss), t@(F _ ts)) =
  conj [conj [ varGtSPO table (s, ti) | ti <- ts ],
        disj [encodeGtB (s, t), 
              conj [encodeGeqB (s, t), encodeGtSPOLex table ss ts]]]
encodeGtSPO23 _ _ = error "encodeGtSPO2"

-- ss >_spo^lex ts
encodeGtSPOLex :: Table -> [Term] -> [Term] -> Formula
encodeGtSPOLex _ []      _  = Or []
encodeGtSPOLex _ (_ : _) [] = And []
encodeGtSPOLex table (s : ss) (t : ts) =
  disj [varGtSPO table (s, t),
        conj [varGeqSPO table (s, t), encodeGtSPOLex table ss ts]]


-- ss >=_spo^mul ts
encodeGtSPOMul :: Table -> [Term] -> [Term] -> Formula
encodeGtSPOMul table ss ts
  | ss' == [] = Or []
  | otherwise = 
    conj [ disj [ encodeGtSPO table (s, t) | s <- ss' ] | t <- ts' ]
  where
    ss' = ss \\ ts
    ts' = ts \\ ss

-- [(x1,x2),(x1,x3),...,(x1,xn),
--          (x2,x3),....(x2,xn),
--                          ...,
--                 (x_{n-1},xn)]
triangle :: [a] -> [(a,a)]
triangle []       = []
triangle (x : xs) = [ (x, y) | y <- xs ] ++ triangle xs

encodeLinearGWPO :: TRS -> Formula
encodeLinearGWPO trs =
  conj ([ varGtSPO table rule | rule <- trs ] ++
        [ conj [SMT.geq (varF0 f) (Val 0), 
                SMT.geq (varF0 (f ++ "#")) (Val 0)]
        | f <- fs ] ++
        [ encodeGeqA rule | rule <- trs ] ++
        [ implies (varGeqSPO table rule) (encodeGeqSPO table rule) | rule <- table ] ++
        [ implies (varGtSPO  table rule) (encodeGtSPO  table rule) | rule <- table ])
  where 
    fs = TRS.functions trs
    table = nub [ (s, t) | (l, r) <- trs, s <- subterms l, t <- subterms r ]

-- Encoding functions for WPO.

encodeLinearWPO :: TRS -> Formula
encodeLinearWPO trs =
  conj (encodeLinearGWPO trs : 
        [ SMT.eq (varF0 f) (varF0 (f ++ "#")) | (f, _) <- sig ] ++
        [ conj [varFi f i, varFi (f ++ "#") i] | (f, n) <- sig, i <- [1..n] ])
  where
    sig = TRS.signatureOf trs

-- Encoding functions for KBO.

varW0 :: Exp
varW0 = Var "w0"

encodeValidityOfWeight :: Signature -> Formula
encodeValidityOfWeight sig =
  conj (Gt varW0 (Val 0) :
        [ Geq (varF0 f) varW0   | (f, 0) <- sig ] ++
        [ Geq (varF0 f) (Val 0) | (f, n) <- sig, n > 0 ])

encodeAdmissibility1 :: Signature -> String -> Formula
encodeAdmissibility1 sig f =
  disj [Gt (varF0 f) (Val 0),
        conj [ Geq (varPrec f) (varPrec g) | (g, _) <- sig ] ]

encodeAdmissibility :: Signature -> Formula
encodeAdmissibility sig = 
  conj [ encodeAdmissibility1 sig f | (f, 1) <- sig ] 

encodeKBO :: TRS -> Formula
encodeKBO trs =
  conj [encodeLinearWPO trs,
        encodeValidityOfWeight sig,
        encodeAdmissibility sig ]
  where
    sig = TRS.signatureOf trs

-- Decoding functions

decodeInterpretation :: String -> Int -> Model -> Interpretation
decodeInterpretation f n model = (f, (f0, fs))
  where 
    f0 = evalExp model (varF0 f)
    fs = [ if evalFormula model (varFi f i) then 1 else 0 | i <- [1 .. n] ]

decodeAlgebra :: TRS -> Model -> Algebra
decodeAlgebra trs model =
  [ decodeInterpretation g n model 
  | (f, n) <- TRS.signatureOf trs, g <- [f, f ++ "#"] ]

decodePrecedence :: TRS -> Model -> Precedence
decodePrecedence trs model =
  [ (f, evalExp model (varPrec f)) | f <- TRS.functions trs ]

decode :: TRS -> Model -> Parameter
decode trs model =
  (decodeAlgebra trs model,
   decodePrecedence trs model)

checkGtMSPO :: Parameter -> TRS -> IO ()
checkGtMSPO _ [] = return ()
checkGtMSPO param (rule : trs)
  | gtMSPO param rule = checkGtMSPO param trs
  | otherwise       = 
    error ("checkGtMSPO: " ++ showRule rule ++ " is not oriented.\n" ++ show (Linear param) ++ "\n")

solve :: (TRS -> Formula) -> String -> TRS -> IO (Maybe Order)
solve encode tool trs = do
  result <- sat tool (encode trs)
  case result of
    Nothing    -> return Nothing
    Just model -> do
      let param = decode trs model
      checkGtMSPO param trs
      return (Just (Linear param))

linearGWPO :: String -> TRS -> IO (Maybe Order)
linearGWPO tool trs = solve encodeLinearGWPO tool trs

linearWPO :: String -> TRS -> IO (Maybe Order)
linearWPO tool trs = solve encodeLinearWPO tool trs

kbo :: String -> TRS -> IO (Maybe Order)
kbo tool trs = solve encodeKBO tool trs
