module MaxPlusGWPO where

import Data.List
import Term
import Rule
import TRS
import SMT
import MaxPlus
import SymbolicMaxPlus as SMaxPlus

-- (f0, [(f1, f'1), ..., (fn, f'n)]) represents
-- max{ f0, f1 + f'1 x_1, ..., fn + f'n x_n }
type MaxPlusParam = (Int, [(Int, Int)])

-- f(x1,...,xn) = max{ f0, f1 + f'1 x_1, ..., fn + f'n x_n }
type Interpretation = (String, MaxPlusParam)

type Algebra = [Interpretation]

type Precedence = [(String, Int)]

type Parameter = (Algebra, Precedence)

data Order = MaxPlus Parameter

type Table = [(Term,Term)]

showMonomial :: (Int, String) -> String
showMonomial (0, _) = "0"
showMonomial (1, x) = x
showMonomial (n, x) = show n ++ " " ++ x

-- pretty printer for f + f'x
showLinearPolynomial :: (Int, Int, String) -> String
showLinearPolynomial (f, 0, _x) = show f
showLinearPolynomial (0, 1, x) = x
showLinearPolynomial (f, 1, x) = show f ++ " + " ++ x
showLinearPolynomial (0, f', x) = show f' ++ x
showLinearPolynomial (f, f', x) = show f ++ " + " ++ show f' ++ " " ++ x

-- can be improved
showMaxPlus :: MaxPlusParam -> String
showMaxPlus (a0, []) = show a0
showMaxPlus (a0, as) = "max{" ++ (intercalate ", " ms) ++ "}"
  where
    m1n = [ showLinearPolynomial (ai, ai', "x" ++ show i)
          | (i, (ai, ai')) <- zip [ 1 :: Int ..] as]
    ms = show a0 : m1n

showInterpretation :: Interpretation -> String
showInterpretation (f, (a0, [])) =
  f ++ "_A = " ++ show a0
showInterpretation (f, mp@(_, as)) =
  f ++ "_A(" ++ xs ++ ") = " ++ showMaxPlus mp
  where
    xs = intercalate "," [ "x" ++ show i | i <- [1 .. length as] ]

showAlgebra :: Algebra -> String
showAlgebra interpretations =
  "max/plus interpretations on N:\n" ++
  -- show interpretations ++ -- DEBUG
  "\n" ++
  unlines 
    [ "  " ++ showInterpretation interpretation 
    | interpretation <- interpretations ] ++ "\n"

showPrecedence' :: Precedence -> String
showPrecedence' []  = error "showPrecedence"
showPrecedence' [(f,_)] = f
showPrecedence' ((f,m) : p@((_,n) : _))
  | m == n = f ++ " = " ++ showPrecedence' p
  | m >  n = f ++ " > " ++ showPrecedence' p
  | otherwise = error "showPrecedence"

showPrecedence :: Precedence -> String
showPrecedence []  =  
  "precedence:\n\n" ++ "    empty\n\n"
showPrecedence [_] = 
  "precedence:\n\n" ++ "    empty\n\n"
showPrecedence a = 
  "precedence:\n\n" ++
  "  " ++ showPrecedence' (sortBy f a) ++ "\n\n"
  where f (_,m) (_,n) = compare n m

showParameter :: Parameter -> String
showParameter (a, p) =
  showAlgebra a ++
  showPrecedence p

instance Show Order where
  show (MaxPlus param) = showParameter param

interpret :: Algebra -> Term -> MaxPlus
interpret _ (V x) = MaxPlus.var x 
interpret a (F f ts)
  | Just (a0, as) <- lookup f a = 
    MaxPlus.flatten (MaxPlus.num a0 : [ MaxPlus.plus ai (MaxPlus.times ai' (interpret a ti)) | ((ai, ai'), ti) <- zip as ts ])
interpret _ _ = error "interpret"

-- >=_A
geqA :: Algebra -> Rule -> Bool
geqA a (l, r) = MaxPlus.geq (interpret a l) (interpret a r)

-- >_A
gtA :: Algebra -> Rule -> Bool
gtA a (l, r) = MaxPlus.gt (interpret a l) (interpret a r)

-- f >= g
geqPrec :: Precedence -> Rule -> Bool
geqPrec p (F f _, F g _)
  | m <- lookup f p, n <- lookup g p = m >= n
geqPrec _ _ = error "geqPrec"

-- f > g
gtPrec :: Precedence -> Rule -> Bool
gtPrec p (F f _, F g _)
  | m <- lookup f p, n <- lookup g p = m > n
gtPrec _ _ = error "geqPrec"

-- s \sqsupsetsim t
geqB :: Parameter -> Rule -> Bool
geqB (a, p) rule = 
  gtA a (Rule.mark rule) ||
  (geqA a (Rule.mark rule) && geqPrec p rule)

-- s \sqsupset t
gtB :: Parameter -> Rule -> Bool
gtB (a,p) rule = 
  gtA a (Rule.mark rule) ||
  (geqA a (Rule.mark rule) && gtPrec p rule)

geqSPO :: Parameter -> Rule -> Bool
geqSPO _ (V x, V y) = x == y
geqSPO _ (V _, _) = False
geqSPO param (F _ ss, t)
  | any (\si -> geqSPO param (si, t)) ss = True
geqSPO _ (F _ _, V _) = False
geqSPO param (s@(F _ ss), t@(F _ ts)) =
  all (\ti -> gtSPO param (s, ti)) ts &&
  (gtB param (s, t) || 
   (geqB param (s, t) && geqSPOLex param ss ts))

geqSPOLex :: Parameter -> [Term] -> [Term] -> Bool
geqSPOLex _ _  []      = True
geqSPOLex _ [] (_ : _) = False
geqSPOLex param (s : ss) (t : ts) =
  gtSPO param (s, t) || 
  (geqSPO param (s, t) && geqSPOLex param ss ts)

gtSPO :: Parameter -> Rule -> Bool
gtSPO _ (V _, _) = False
gtSPO param (F _ ss, t)
  | any (\si -> geqSPO param (si, t)) ss = True
gtSPO _ (F _ _, V _) = False
gtSPO param (s@(F _ ss), t@(F _ ts)) =
  all (\ti -> gtSPO param (s, ti)) ts &&
  (gtB param (s, t) || 
   (geqB param (s, t) && gtSPOLex param ss ts))

gtSPOLex :: Parameter -> [Term] -> [Term] -> Bool
gtSPOLex _ [] _        = False
gtSPOLex _ (_ : _)  [] = True
gtSPOLex param (s : ss) (t : ts) =
  gtSPO param (s, t) || 
  (geqSPO param (s, t) && gtSPOLex param ss ts)

gtSPOMul :: Parameter -> [Term] -> [Term] -> Bool
gtSPOMul param ss ts =
  ss'/= [] &&
  and [ or [ gtSPO param (s, t) | s <- ss' ] | t <- ts' ]
  where 
    ss' = ss \\ ts
    ts' = ts \\ ss

geqMSPO :: Parameter -> Rule -> Bool
geqMSPO param@(a,_) rule = geqA a rule && geqSPO param rule

gtMSPO :: Parameter -> Rule -> Bool
gtMSPO param@(a,_) rule = geqA a rule && gtSPO param rule

-- Encoding functions.

varF0 :: String -> Exp
varF0 f = Var ("a0_" ++ f) 

varFi :: String -> Int -> Exp
varFi f i = Var ("a_" ++ show i ++ "_" ++ f) 

varF'i :: String -> Int -> Formula
varF'i f i = FVar ("b_" ++ show i ++ "_" ++ f) 

varPrec :: String -> Exp
varPrec f = Var ("p_" ++ f)

varGeqSPO :: Table -> Rule -> Formula
varGeqSPO table rule
  | Just i <- elemIndex rule table = FVar ("geqSPO_" ++ show i)
  | otherwise                      = error "varGeqSPO"

varGtSPO :: Table -> Rule -> Formula
varGtSPO table rule
  | Just i <- elemIndex rule table = FVar ("gtSPO_" ++ show i)
  | otherwise                      = error "varGtSPO"

symbolicMaxPlus :: Term -> SymbolicMaxPlus
symbolicMaxPlus (V x) = SMaxPlus.var x
symbolicMaxPlus (F f ts) =
   SMaxPlus.flatten (SMaxPlus.const (varF0 f) : [ multiplyAdd f i ti | (i, ti) <- zip [1..] ts ])
  where
    multiplyAdd fname i ti = SMaxPlus.addConstant (varFi fname i) (SMaxPlus.mult (varF'i fname i) (symbolicMaxPlus ti))

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA (l, r) = SMaxPlus.encodeGeq (symbolicMaxPlus l) (symbolicMaxPlus r)

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA (l, r) = SMaxPlus.encodeGt (symbolicMaxPlus l) (symbolicMaxPlus r)

-- f >= g
encodeGeqPrec :: Rule -> Formula
encodeGeqPrec (F f _, F g _) = SMT.geq (varPrec f) (varPrec g)
encodeGeqPrec _ = error "encodeGeqPrec"

-- f > g
encodeGtPrec :: Rule -> Formula
encodeGtPrec (F f _, F g _) = SMT.gt (varPrec f) (varPrec g)
encodeGtPrec _ = error "encodeGtPrec"

-- f(s1,...,sm) >=_B g(t1,...,tn) if 
--  - s^# >_A t^#, or
--  - s^# >_A t^# and f >= g
encodeGeqB :: Rule -> Formula
encodeGeqB rule = 
  disj [encodeGtA (Rule.mark rule),
        conj [encodeGeqA (Rule.mark rule), encodeGeqPrec rule]]

-- f(s1,...,sm) >_B g(t1,...,tn) if 
--  - s^# >_A t^#, or
--  - s^# >=_A t^# and f >= g
encodeGtB :: Rule -> Formula
encodeGtB rule = 
  disj [encodeGtA (Rule.mark rule),
        conj [encodeGeqA (Rule.mark rule), encodeGtPrec rule]]

-- s >=_spo t
encodeGeqSPO :: Table -> Rule -> Formula
encodeGeqSPO _ (V x, V y)
  | x == y    = And []
encodeGeqSPO _ (V _, _) = Or []
encodeGeqSPO table rule@(F _ _, V _) =
  encodeGeqSPO1 table rule
encodeGeqSPO table rule@(F _ _, F _ _) =
  disj [encodeGeqSPO1 table rule, encodeGeqSPO23 table rule]


encodeGeqSPO1 :: Table -> Rule -> Formula
encodeGeqSPO1 table (F _ ss, t) =
  disj [ varGeqSPO table (si, t) | si <- ss ]
encodeGeqSPO1 _ _ = error "encodeGeqSPO1"

encodeGeqSPO23 :: Table -> Rule -> Formula
encodeGeqSPO23 table (s@(F _ ss), t@(F _ ts)) =
  conj [conj [ varGtSPO table (s, ti) | ti <- ts ],
        disj [encodeGtB (s, t), 
              conj [encodeGeqB (s, t), encodeGeqSPOLex table ss ts]]]
encodeGeqSPO23 _ _ = error "encodeGeqSPO2"

-- ss >=_spo^lex ts
encodeGeqSPOLex :: Table -> [Term] -> [Term] -> Formula
encodeGeqSPOLex _ _  []      = And []
encodeGeqSPOLex _ [] (_ : _) = Or []
encodeGeqSPOLex table (s : ss) (t : ts) =
  disj [varGtSPO table (s, t),
        conj [varGeqSPO table (s, t), encodeGeqSPOLex table ss ts]]

-- ss >=_spo^mul ts
encodeGeqSPOMul :: Table -> [Term] -> [Term] -> Formula
encodeGeqSPOMul table ss ts =
  conj [ disj [ encodeGeqSPO table (s, t) | s <- ss \\ ts ] 
       | t <- ts \\ ss ]

-- s >_spo t
encodeGtSPO :: Table -> Rule -> Formula
encodeGtSPO _ (V _, _) = Or []
encodeGtSPO table (s@(F _ _), t@(V _)) =
  encodeGtSPO1 table (s, t)
encodeGtSPO table rule@(F _ _, F _ _) =
  disj [encodeGtSPO1 table rule, encodeGtSPO23 table rule]

encodeGtSPO1 :: Table -> Rule -> Formula
encodeGtSPO1 table (F _ ss, t) =
  disj [ varGeqSPO table (si, t) | si <- ss ]
encodeGtSPO1 _ _ = error "encodeGtSPO1"

encodeGtSPO23 :: Table -> Rule -> Formula
encodeGtSPO23 table (s@(F _ ss), t@(F _ ts)) =
  conj [conj [ varGtSPO table (s, ti) | ti <- ts ],
        disj [encodeGtB (s, t), 
              conj [encodeGeqB (s, t), encodeGtSPOLex table ss ts]]]
encodeGtSPO23 _ _ = error "encodeGtSPO2"

-- ss >_spo^lex ts
encodeGtSPOLex :: Table -> [Term] -> [Term] -> Formula
encodeGtSPOLex _ []      _  = Or []
encodeGtSPOLex _ (_ : _) [] = And []
encodeGtSPOLex table (s : ss) (t : ts) =
  disj [varGtSPO table (s, t),
        conj [varGeqSPO table (s, t), encodeGtSPOLex table ss ts]]


-- ss >=_spo^mul ts
encodeGtSPOMul :: Table -> [Term] -> [Term] -> Formula
encodeGtSPOMul table ss ts
  | ss' == [] = Or []
  | otherwise = 
    conj [ disj [ encodeGtSPO table (s, t) | s <- ss' ] | t <- ts' ]
  where
    ss' = ss \\ ts
    ts' = ts \\ ss

-- [(x1,x2),(x1,x3),...,(x1,xn),
--          (x2,x3),....(x2,xn),
--                          ...,
--                 (x_{n-1},xn)]
triangle :: [a] -> [(a,a)]
triangle []       = []
triangle (x : xs) = [ (x, y) | y <- xs ] ++ triangle xs

-- f_j can be minus
encodeMaxPlusGWPOZ :: TRS -> Formula
encodeMaxPlusGWPOZ trs =
  conj ([ varGtSPO table rule | rule <- trs ] ++
        [ conj [SMT.geq (varF0 f) (Val 0),
                SMT.geq (varF0 (f ++ "#")) (Val 0)]
        | f <- TRS.functions trs  ] ++
        [ encodeGeqA rule | rule <- trs ] ++
        [ implies (varGeqSPO table rule) (encodeGeqSPO table rule) | rule <- table ] ++
        [ implies (varGtSPO  table rule) (encodeGtSPO  table rule) | rule <- table ])
  where 
    table = nub [ (s, t) | (l, r) <- trs, s <- subterms l, t <- subterms r ]

-- f_j >= 0
encodeMaxPlusGWPO :: TRS -> Formula
encodeMaxPlusGWPO trs =
  conj (encodeMaxPlusGWPOZ trs :
        [ conj [SMT.geq (varFi f i) (Val 0),
                SMT.geq (varFi (f ++ "#") i) (Val 0)]
        | (f, n) <- TRS.signatureOf trs, i <- [1..n]])

encodeMaxPlusWPO :: TRS -> Formula
encodeMaxPlusWPO trs =
  conj (encodeMaxPlusGWPO trs : 
        [ SMT.eq (varF0 f) (varF0 (f ++ "#")) | (f, _) <- sig ] ++
        [ SMT.eq (varFi f i) (varFi (f ++ "#") i) | (f, n) <- sig, i <- [1..n] ] ++
        [ conj [varF'i f i, varF'i (f ++ "#") i] | (f, n) <- sig, i <- [1..n] ])
  where
    sig = TRS.signatureOf trs

encodeLPO :: TRS -> Formula
encodeLPO trs =
  conj (encodeMaxPlusWPO trs : 
        [ SMT.eq (varF0 f) (Val 0) | (f, _) <- sig ] ++
        [ SMT.eq (varFi f i) (Val 0) | (f, n) <- sig, i <- [1..n] ])
  where
    sig = TRS.signatureOf trs

decodeInterpretation :: String -> Int -> Model -> Interpretation
decodeInterpretation f n model = (f, (f0, ff's))
  where 
    f0 = evalExp model (varF0 f)
    evalf fname i = evalExp model (varFi fname i)
    evalf' fname i = if evalFormula model (varF'i fname i) then 1 else 0
    ff's = [ (evalf f i, evalf' f i) | i <- [1 .. n] ]

decodeAlgebra :: TRS -> Model -> Algebra
decodeAlgebra trs model =
  [ decodeInterpretation g n model 
  | (f, n) <- TRS.signatureOf trs, g <- [f, f ++ "#"] ]

decodePrecedence :: TRS -> Model -> Precedence
decodePrecedence trs model =
  [ (f, evalExp model (varPrec f)) | f <- TRS.functions trs ]

decode :: TRS -> Model -> Parameter
decode trs model =
  (decodeAlgebra trs model,
   decodePrecedence trs model)

checkGtMSPO :: Parameter -> TRS -> IO ()
checkGtMSPO _ [] = return ()
checkGtMSPO param (rule : trs)
  | gtMSPO param rule = checkGtMSPO param trs
  | otherwise       = 
    error ("checkGtMSPO: " ++ showRule rule ++ " is not oriented.\n" ++ show (MaxPlus param) ++ "\n")

solve :: (TRS -> Formula) -> String -> TRS -> IO (Maybe Order)
solve encode tool trs = do
  result <- sat tool (encode trs)
  case result of
    Nothing    -> return Nothing
    Just model -> do
      let param = decode trs model
      checkGtMSPO param trs
      return (Just (MaxPlus param))

maxPlusGWPO :: String -> TRS -> IO (Maybe Order)
maxPlusGWPO tool trs = solve encodeMaxPlusGWPO tool trs

maxPlusGWPOZ :: String -> TRS -> IO (Maybe Order)
maxPlusGWPOZ tool trs = solve encodeMaxPlusGWPOZ tool trs

maxPlusWPO :: String -> TRS -> IO (Maybe Order)
maxPlusWPO tool trs = solve encodeMaxPlusWPO tool trs

lpo :: String -> TRS -> IO (Maybe Order)
lpo tool trs = solve encodeLPO tool trs
